package be.tryalogue;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.TimerTask;

import android.app.Activity;
import android.content.Context;
import android.content.Intent;
import android.content.pm.PackageManager;
import android.content.pm.ResolveInfo;
import android.media.AudioManager;
import android.os.Bundle;
import android.os.Handler;
import android.speech.RecognitionListener;
import android.speech.RecognizerIntent;
import android.speech.SpeechRecognizer;
import android.speech.tts.TextToSpeech;
import android.speech.tts.TextToSpeech.OnInitListener;
import android.speech.tts.TextToSpeech.OnUtteranceCompletedListener;
import android.util.Log;
import android.view.Gravity;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.Window;
import android.view.WindowManager;
import android.widget.Toast;

public class TryalogueActivity extends Activity implements OnClickListener, OnInitListener {

    private static final String TAG = "Tryalogue";
    
    private static final int TTS_CHECK_CODE = 4321;

    private TextToSpeech textToSpeech;
    
    private SpeechRecognizer recognizer;
    
    private Intent recognizerIntent;
    
    private OscilloscopeAnimation animation;
    
    String defaultTexts[] = {
    		"I am sitting in a room different from the one you are in now. I am generating the sound of a speaking voice.", 
    		"I am going to do text to speech followed by speech recognition and repeat that again and again", 
    		"It will repeat until the speech recoginition fails so that any semblance of my message is destroyed.",
    		"What you will hear, then, are is an interpretation of speech by a speech recognition engine.",
    		"I regard this activity not so much as a demonstration of an algorithmic problem.",
    		"It is more a way to expose an irregular internal dialog."};
    
    int currentIndex;
    
    private String text;
    
    /**
     * Called with the activity is first created.
     */
    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        //Remove title bar
        this.requestWindowFeature(Window.FEATURE_NO_TITLE);

        //Remove notification bar
        this.getWindow().setFlags(WindowManager.LayoutParams.FLAG_FULLSCREEN, WindowManager.LayoutParams.FLAG_FULLSCREEN);

        // Inflate our UI from its XML layout description.
        setContentView(R.layout.voice_recognition);

        checkTextToSpeechEngine();
        checkSpeechRecognitionEngine();
        
        recognizerIntent = new Intent(RecognizerIntent.ACTION_RECOGNIZE_SPEECH);
        recognizerIntent.putExtra(RecognizerIntent.EXTRA_LANGUAGE_MODEL,RecognizerIntent.LANGUAGE_MODEL_FREE_FORM);
        recognizerIntent.putExtra(RecognizerIntent.EXTRA_CALLING_PACKAGE, "be.tryalogue");

        recognizer = SpeechRecognizer.createSpeechRecognizer(this.getApplicationContext());
        
        recognizer.setRecognitionListener(listener);
        
        animation = (OscilloscopeAnimation) findViewById(R.id.animation);
        animation.setOnClickListener(this);
        
        text = defaultTexts[currentIndex];
        currentIndex = (currentIndex + 1 ) % defaultTexts.length;
        
        showToast("Tap to start");
       
    }
    
    @Override
    public void onPause() {
    	//this.textToSpeech.stop();
    	//this.recognizer.stopListening();
    	super.onPause();
    }
    
    @Override
    public void onStop(){
    	//this.textToSpeech.stop();
    	//this.recognizer.stopListening();
    	super.onStop();
    }
    
    Handler restartHandler = new Handler();
    
    TimerTask restartTask = new TimerTask() {
		@Override
		public void run() {
			restartRecognizer();
		}
	};
    
    RecognitionListener listener = new RecognitionListener() {
        @Override
        public void onResults(Bundle results) {
            ArrayList<String> voiceResults = results.getStringArrayList(SpeechRecognizer.RESULTS_RECOGNITION);
            if (voiceResults == null) {
                Log.e(TAG, "No voice results");
                text = "I am sitting in a room different from the one you are in now, I am generating the sound of a speaking voice.";// and I am going to do text to speech followed by speech recognition again and again until any semblance to the original sentence is destroyed.";
            } else {
                String newText = getLongestMatch(voiceResults);
                if(newText.equalsIgnoreCase(text) || newText.length() < 40){
                	Log.i(TAG,"Skipping to next text");
                	text = defaultTexts[currentIndex];
                    currentIndex = (currentIndex + 1 ) % defaultTexts.length;
                } else {
                	text = newText;
                }
                Log.i(TAG,"Match: " + text);
            }
        }

        private String getLongestMatch(ArrayList<String> voiceResults) {
			String word = voiceResults.get(0);
			for(int i = 0; i < voiceResults.size();i++){
				String current = voiceResults.get(i);
				if(current.length() > word.length()){
					word = current;
				}
			}
			return word;
		}

		@Override
        public void onReadyForSpeech(Bundle params) {
            Log.d(TAG, "Ready for speech");
            //provide a hash so onUtterancefinished callback is called (This api sucks)
            HashMap<String, String> aHash = new HashMap<String, String>();
            aHash.put(TextToSpeech.Engine.KEY_PARAM_STREAM,String.valueOf(AudioManager.STREAM_ALARM));
            aHash.put(TextToSpeech.Engine.KEY_PARAM_UTTERANCE_ID,"Random ID");
            textToSpeech.speak(text, TextToSpeech.QUEUE_FLUSH, aHash);
            showToast(text);
	        Log.i(TAG, "Speak: " + text);
        }
		
		long lastRestart = 0L;

        @Override
        public void onError(int error) {
            Log.d(TAG,"Error listening for speech: " + error);
            animation.setRmsDB(0);
            if(error == SpeechRecognizer.ERROR_RECOGNIZER_BUSY &&  System.currentTimeMillis() - lastRestart > 2200 ){
            	lastRestart = System.currentTimeMillis();
            	Log.d(TAG,"Recognizer was still busy, restart in 1300ms: " + error);
            	restartHandler.removeCallbacks(restartTask);
                restartHandler.postDelayed(restartTask, 2000);
            } else if (  error == SpeechRecognizer.ERROR_NO_MATCH ){
            	text = defaultTexts[currentIndex];
                currentIndex = (currentIndex + 1 ) % defaultTexts.length;
                Log.d(TAG,"No match found skip to next text, restart in 1300ms: " + error);
            	restartHandler.removeCallbacks(restartTask);
                restartHandler.postDelayed(restartTask, 2000);
            }
        }

        @Override
        public void onBeginningOfSpeech() {
            Log.d(TAG, "Speech started");
        }

        @Override
        public void onBufferReceived(byte[] buffer) {
           animation.setBytes(buffer);
        }

        @Override
        public void onEndOfSpeech() {
        	Log.d(TAG, "Speech ended");
        	animation.setRmsDB(0);
        	//restart the loop on the main thread in a second
            restartHandler.removeCallbacks(restartTask);
            restartHandler.postDelayed(restartTask, 2000);
        }

        @Override
        public void onEvent(int eventType, Bundle params) {
            // TODO Auto-generated method stub
        }

        @Override
        public void onPartialResults(Bundle partialResults) {
            // TODO Auto-generated method stub
        }

        @Override
        public void onRmsChanged(float rmsdB) {
            animation.setRmsDB(rmsdB);
        }
    };
    
    private void checkTextToSpeechEngine(){
    	Intent checkIntent = new Intent();
    	checkIntent.setAction(TextToSpeech.Engine.ACTION_CHECK_TTS_DATA);
    	startActivityForResult(checkIntent, TTS_CHECK_CODE);
    }
    
    private void checkSpeechRecognitionEngine(){
    	// Check to see if a recognition activity is present
        PackageManager pm = getPackageManager();
        List<ResolveInfo> activities = pm.queryIntentActivities(new Intent(RecognizerIntent.ACTION_RECOGNIZE_SPEECH), 0);
        if (activities.size() == 0) {
            showToast("Speech recognition engine not not present.\nPlease install Voice Search");
        }
    }

    /**
     * Handle the click on the start recognition button.
     */
    public void onClick(View v) {
    	startVoiceRecognitionActivity();
    }

    /**
     * Fire an intent to start the speech recognition activity.
     */
    protected void startVoiceRecognitionActivity() {
    	recognizer.startListening(recognizerIntent);
    }

    
    /**
     * Handle the results from the check voice data activity.
     */
    @Override
    protected void onActivityResult(int requestCode, int resultCode, Intent data) {
       if(requestCode == TTS_CHECK_CODE){
        	 if (resultCode == TextToSpeech.Engine.CHECK_VOICE_DATA_PASS) {
                 // success, create the TTS instance
                 textToSpeech = new TextToSpeech(this, this);
                 //set the max volume
                 AudioManager am = (AudioManager)getSystemService(Context.AUDIO_SERVICE);
                 int amStreamMusicMaxVol = am.getStreamMaxVolume(AudioManager.STREAM_SYSTEM);
                 am.setStreamVolume(AudioManager.STREAM_SYSTEM, amStreamMusicMaxVol, 0);
             } else {
                 // missing data, install it
            	 Log.i(TAG,"Install text to speech data.");
                 Intent installIntent = new Intent();
                 installIntent.setAction(TextToSpeech.Engine.ACTION_INSTALL_TTS_DATA);
                 startActivity(installIntent);
             }
        }
        super.onActivityResult(requestCode, resultCode, data);
    }
    
    private OnUtteranceCompletedListener utteranceCompleteListener = new OnUtteranceCompletedListener() {
		@Override
		public void onUtteranceCompleted(String utteranceId) {
			Log.i(TAG, "Utterance Complete");
		}
	};
	
	private void restartRecognizer(){
		recognizer.stopListening();
		startVoiceRecognitionActivity();
	}

    /**
     * Shows a toast message
     * @param text
     */
    private void showToast(String text) {
        Toast t = Toast.makeText(TryalogueActivity.this, text,Toast.LENGTH_SHORT);
        t.setGravity(Gravity.CENTER,0,-120);
        t.show();
    }

	@Override
	public void onInit(int arg0) {
		// TODO Auto-generated method stub
        textToSpeech.setOnUtteranceCompletedListener(utteranceCompleteListener);
		Log.i(TAG, "Text to speech engine initialized.");
	}
}